
const B3LayoutExt = () => {
  const ExtensionId = 'Viewing.Extension.B3L'
  let UseObjects = true;
  let self;
  let pointColor = 0xff0000;
  let intersectColor = 0x00D66B;
  let selectColor = 0xbfd1e5;
  let pivot;
  let offset;
  let lastPos;
  let lastTarget;
  let lastWasPerspective;
  let camParamsFinal;
  let in2d = false;
  let sectionActive = false;
  let obsPoint = null;
  let hiddenDbIds = [];
  let sectionPlanes = [];
  const fullPropFilter = ['PointNumber', 'POINTNUMBER', 'PointDescription', 'DESC', 'DESCRIPTION', 'Description', 'PointRole', 'POINTROLE', 'Role', 'POINTTYPE', 'PointType', 'Type Name', 'SYSTEM', 'ALL_MODEL_TYPE_NAME', 'LcRevitData_Element:LcRevitPropertyElementType'];
  const partPropFilter = ['PointNumber', 'POINTNUMBER', 'POINTTYPE', 'PointRole', 'POINTROLE'];



  class B3Layout
    extends Autodesk.Viewing.Extension {

    constructor (viewer, options) {
      super(viewer, options);
      self = this;
      self.viewer = viewer;
      self.camera = self.viewer.navigation.getCamera();
      self.mouse = new THREE.Vector2();
      self.intersected = null;
      self.raycaster = new THREE.Raycaster();
      self.raycaster.params.PointCloud.threshold = .01;
      self.highlightObj = self.GetSphere(intersectColor, 0, 0, 0);
      self.selectObj = self.GetSphere(selectColor, 0, 0, 0);
      self.highlightObj.visible = false;
      self.selectObj.visible = false;
      self.APLpts = [];
      self.lAPLpts = [];
      self.LocalPtsFound = 0;
      //Since we are in headless mode, need to add in progressbar if we want it
      //TODO add check if in headless mode
      self.progressbar = new Autodesk.Viewing.Private.ProgressBar(viewer.container);
      self.viewer.addEventListener(Autodesk.Viewing.PROGRESS_UPDATE_EVENT, (event) => { if (event.percent >= 0) { self.progressbar.setPercent(event.percent); }});
      self.setProgressBarLocation();
    }

    load () {
      return true;
    }

    unload () {
      return true;
    }

    worldToClient(x,y,z) {
        let pos = new THREE.Vector3(x,y,z);
        return self.viewer.worldToClient(pos);
    }

    clientToWorld(x,y,z) {
        return self.viewer.clientToWorld(x,y,z);    //see pageToModel, viewportToClient
    }

    getPageToModelTransform() {
        var b = self.viewer.model.getData();
        return self.viewer.model.getData().pageToModelTransform();
    }

    getGlobalOffset() {
        self.offset = self.viewer.model?self.viewer.model.getData().globalOffset:{x:0,y:0,z:0};
        return self.offset;
    }

    getUnits() {
        return self.viewer.model?self.viewer.model.getUnitString():"Error";
//        return self.viewer.model?self.viewer.model.getMetadata("distance unit","value",null):"Error";
    }

    setProgressBarLocation (fromBottom) {
      if(!fromBottom)
        fromBottom = 60;
      jss.set('.progressbg', {
          'bottom': `${fromBottom}px`
        });
    }

    setIssueSize (size) {
      var markerProps;
      var markerSelProps;
      if(!size || size == 100){ //Default
        markerProps = {
          'width': '22px',
          'height': '22px',
          'left': '0px',
          'top': '0px',
          'border': '4px solid rgba(255, 255, 255, 0.95)',
          'background-clip': '',
          'border-radius' : '',
          'box-shadow' : ''
        }
        markerSelProps = {
          'width': '28px',
          'height': '28px',
          'left': '0px',
          'top': '0px',
          'border': '5px solid rgba(255, 255, 255, 0.95)',
          'background-clip': '',
          'border-radius' : '',
          'box-shadow' : ''
        }
      }else if(size === 25){ //25%
        markerProps = {
          'width': '7px',
          'height': '7px',
          'left': '-9px',
          'top': '-9px',
          'border': '20px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 1px inset'
        }
        markerSelProps = {
          'width': '12px',
          'height': '12px',
          'left': '-12px',
          'top': '-12px',
          'border': '25px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 2px inset'
        }
      }else if(size === 37){ //37%
        markerProps = {
          'width': '10px',
          'height': '10px',
          'left': '-6px',
          'top': '-6px',
          'border': '16px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 2px inset'
        }
        markerSelProps = {
          'width': '17px',
          'height': '17px',
          'left': '-6px',
          'top': '-6px',
          'border': '16px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 3px inset'
        }
      }else if(size === 50){ //50%
        markerProps = {
          'width': '12px',
          'height': '12px',
          'left': '-6px',
          'top': '-6px',
          'border': '14px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 2px inset'
        }
        markerSelProps = {
          'width': '19px',
          'height': '19px',
          'left': '-5px',
          'top': '-5px',
          'border': '14px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 3px inset'
        }
      }else if(size === 75){ //75%
        markerProps = {
          'width': '19px',
          'height': '19px',
          'left': '-2px',
          'top': '-2px',
          'border': '7px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 3px inset'
        }
        markerSelProps = {
          'width': '26px',
          'height': '26px',
          'left': '-2px',
          'top': '-2px',
          'border': '7px solid rgba(255, 255, 255, 0)',
          'background-clip': 'content-box',
          'border-radius' : '30px',
          'box-shadow' : 'rgba(255, 255, 255, 0.95) 0px 0px 0px 4px inset'
        }
      }else if(size === 150){ //150%
        markerProps = {
          'width': '32px',
          'height': '32px',
          'left': '-7px',
          'top': '-7px',
          'border': '6px solid rgba(255, 255, 255, 0.95)',
          'background-clip': '',
          'border-radius' : '30px',
          'box-shadow' : ''
        }
        markerSelProps = {
          'width': '40px',
          'height': '40px',
          'left': '-8px',
          'top': '-8px',
          'border': '7px solid rgba(255, 255, 255, 0.95)',
          'background-clip': '',
          'border-radius' : '30px',
          'box-shadow' : ''
        }
      }
      jss.set('.pushpin-billboard-marker', 
      //jss.set('.rfi-billboard-marker',
      markerProps);
      jss.set('.pushpin-billboard-marker.selected',
      //jss.set('.rfi-billboard-marker.selected', 
      markerSelProps);
    }

    setIssueColor (rgb) {
      jss.set('.pushpin-billboard-marker', {
      //jss.set('.rfi-billboard-marker', {
        'background-color': rgb  //'rgb(255, 12, 24)'
      });
    }

    showMem() {
        var memInfo = self.viewer.getMemoryInfo();
        if (memInfo) {
            console.log("mem limit:" + memInfo.limit);
            console.log("effective:" + memInfo.effectiveLimit);
            console.log("memloaded:" + memInfo.loaded);
        } else console.log("fits within memory");
    }

    getMeta() {
        return self.viewer.svf?self.viewer.svf.metadata["custom values"]:"Error";
    }

    fitBounds(xmin, ymin, zmin, xmax, ymax, zmax) {
        var c = new THREE.Box3(new THREE.Vector3(xmin, ymin, zmin),new THREE.Vector3(xmax, ymax, zmax));
        return self.viewer.model?self.viewer.navigation.fitBounds(false,c,false):"Error";
    }

    toggle2d() {
        let camera = self.viewer.navigation.getCamera();
        if (self.in2d) {
            self.viewer.navigation.setIsLocked(false);
            if(self.lastWasPerspective===1){self.viewer.navigation.toPerspective();}
            this.viewer.navigation.setTarget(self.lastTarget);
            this.viewer.navigation.setPosition(self.lastPos);
            this.viewer.navigation.setCameraUpVector(self.lastUp);
            self.in2d = false;
        } else {
            self.lastPos = new THREE.Vector3().copy(this.viewer.navigation.getPosition());
            self.lastUp = new THREE.Vector3().copy(this.viewer.navigation.getCameraUpVector());
            self.lastTarget = new THREE.Vector3().copy(this.viewer.navigation.getTarget())
            self.lastWasPerspective = camera.isPerspective?1:0;
            self.viewer.navigation.toOrthographic();
            self.setViewTransform('top');
            self.in2d = true;
        }
     }

    lock2d() {
        self.viewer.navigation.setLockSettings({"pan":true, "zoom":true, "gotoview":true});
        self.viewer.navigation.setIsLocked(true);
    }

    pointToRaycaster (domElement, camera, point) {

      const pointerVector = new THREE.Vector3()
      const pointerDir = new THREE.Vector3()
      const ray = new THREE.Raycaster()
  
      const rect = domElement.getBoundingClientRect()
  
      const x =  ((point.x - rect.left) / rect.width)  * 2 - 1
      const y = -((point.y - rect.top)  / rect.height) * 2 + 1
  
      if (camera.isPerspective) {
  
        pointerVector.set(x, y, 0.5)
  
        pointerVector.unproject(camera)
  
        ray.set(camera.position,
          pointerVector.sub(
            camera.position).normalize())
  
      } else {
  
        pointerVector.set(x, y, -1)
  
        pointerVector.unproject(camera)
  
        pointerDir.set(0, 0, -1)
  
        ray.set(pointerVector,
          pointerDir.transformDirection(
            camera.matrixWorld))
      }
  
      return ray
    }
  
    hideBlocking(event)
    {
      const clientPoint = self.viewer.worldToClient(obsPoint)
  
      //const offset = $(self.viewer.container).offset()
      let offset = { top: 0, left: 0 };
      if ( self.viewer.container.getClientRects().length ) {
        const rect = self.viewer.container.getBoundingClientRect();
        const win = self.viewer.container.ownerDocument.defaultView;
        offset = {
		      top: rect.top + win.pageYOffset,
  		  	left: rect.left + win.pageXOffset
    		};
      }
  
      const rayCaster = self.pointToRaycaster(
        self.viewer.impl.canvas,
        self.viewer.impl.camera, {
          x: clientPoint.x + offset.left,
          y: clientPoint.y + offset.top
        }
      )
      const dist = rayCaster.ray.origin.distanceTo(obsPoint);
      hiddenDbIds.map((item) => {self.viewer.show(item)});
      hiddenDbIds = [];
      let intersections = [];
      const hitTest = self.viewer.impl.rayIntersect(rayCaster.ray, false, null, null, intersections)
  
      if (intersections.length > 0) {
        intersections = intersections.filter(a=>{return a.distance < dist});
        let intDbIds = intersections.map(a=>a.dbId);
        intDbIds = [...new Set(intDbIds)];
        // what to hide as new intersect
        let hideDbIds = intDbIds.filter(a=>{return hiddenDbIds.indexOf(a) === -1});
        hideDbIds.map((item) => {hiddenDbIds.push(item); self.viewer.hide(item)});
      }

    }

    setObsPt(pt) {
      if(pt)
      {
        if(!obsPoint){
          obsPoint = pt;
          self.viewer.addEventListener(Autodesk.Viewing["CAMERA_CHANGE_EVENT"], this.hideBlocking);
        }else{obsPoint = pt;}
        //Force a camera change event to be fired
        self.viewer.navigation.setPosition(self.viewer.navigation.getPosition())
      }else{
        obsPoint = null;
        self.viewer.removeEventListener(Autodesk.Viewing["CAMERA_CHANGE_EVENT"], this.hideBlocking)
        hiddenDbIds.map((item) => {self.viewer.show(item)});
        hiddenDbIds = [];
      }
    }

    rotateN() {
//        let newCamera1 = new THREE.Vector3(self.camera.position.x, self.lastTarget.y, self.camera.position.z);
//        self.viewer.navigation.setView(newCamera1, self.lastTarget);
    }

    toOrtho() {
          self.viewer.navigation.toOrthographic();
    }

    setViewTransform (faceString){
      let camera = self.viewer.navigation.getCamera();
      var worldUp = camera.worldup ? camera.worldup.clone() : camera.up.clone();
      var worldFront =  worldUp.clone().cross(self.getWorldRightVector()).normalize();
      var worldRight = worldFront.clone().cross(worldUp).normalize();

      var camParamsInitial = camera.clone();
      camParamsInitial.center = camera.target ? camera.target.clone() : new THREE.Vector3(0,0,0);
      camParamsInitial.pivot = camera.pivot ? camera.pivot.clone() : camParamsInitial.center.clone();

      camParamsFinal = camera.clone();
      camParamsFinal.center = camera.target ? camera.target.clone() : new THREE.Vector3(0,0,0);
      camParamsFinal.pivot = camera.pivot ? camera.pivot.clone() : camParamsFinal.center.clone();

      // find movement offset based on given boolean flags
      var offset = new THREE.Vector3( 0, 0, 0 );
      if ( faceString.contains('back')){
          offset = offset.add(worldFront);
      }
      if ( faceString.contains('front')){
          offset = offset.sub(worldFront);
      }
      if ( faceString.contains('top')){
          offset = offset.add(worldUp);
      }
      if ( faceString.contains('bottom')){
          offset = offset.sub(worldUp);
      }
      if ( faceString.contains('right')){
          offset = offset.add(worldRight);
      }
      if ( faceString.contains('left')){
          offset = offset.sub(worldRight);
      }
      var upDir = worldUp;

      // view looking at top or bottom chosen
      var test = offset.clone().normalize();

      if ( ( 1.0 - Math.abs(test.dot(worldUp)) ) < Number.MIN_VALUE ) {
          //( offset == worldUp || offset == -worldUp )
          // find the principal view direction other than top/bottom closest to
          // the current view direction and use it as an up vector

          var viewDir = camParamsInitial.center.clone().sub(camParamsInitial.position).normalize();
          var optUpDir = [ worldFront.clone(), worldFront.clone().negate(), worldRight.clone(), worldRight.clone().negate() ];

          // use both view and up vectors for test vector because transitioning from
          // top and bottom views, view direction is the same (but up direction is different)

          var sign = (test.dot(worldUp) > 0.0) ? +1.0 : -1.0; //( offset == worldUp ) ? +1.0 : -1.0;
          var testDir = viewDir.clone().add(camera.up.clone().multiplyScalar(sign)).normalize();

          var optValue = -2.0;

          for ( var i = 0; i < 4; i++ ){
              var value = testDir.dot( optUpDir[i] );

              if ( value > optValue ){
                  optValue = value;
                  upDir = optUpDir[i].multiplyScalar(sign);
              }
          }
      }
      var distFinal = camParamsInitial.center.clone().sub(camera.position).length();
      camParamsFinal.position.copy(camParamsFinal.center.clone().add(offset.multiplyScalar(distFinal/offset.length())));
      camParamsFinal.up.copy(new THREE.Vector3(0,1,0));//upDir);
      self.viewer.applyCamera(camParamsFinal,true);
      self.lock2d();
    }

    getWorldRightVector()
    {
        let camera = self.viewer.navigation.getCamera();
        var vec =  camera.worldup ? camera.worldup.clone() : camera.up.clone();
        if (Math.abs(vec.z) <= Math.abs(vec.y))
        {
            // Cross(Vertical, ZAxis)
            vec.set(vec.y, -vec.x, 0);
        }
        else if (vec.z >= 0)
        {
            // Cross(YAxis, Vertical)
            vec.set(vec.z, 0, -vec.x);
        }
        else
        {
            // Cross(Vertical, YAxis)
            vec.set(-vec.z, 0, vec.x);
        }
        return vec.normalize();
    }
    test(){
      self.viewer.setGhosting(true);
      //const pPt = { "x":-68.03385162353516, "y": -29.22888946533203, "z": 5};
      // const r1Pt = { "x":-68.03385162353516, "y": 8.687776565551758, "z": 5};
      // const r2Pt = { "x":-64.20052337646484, "y": 6.854443073272705, "z": 9};
      const pPt = { "x":-13.826311111450195, "y": -9.387151718139648, "z": -8.794543266296387};
      const r1Pt = { "x": -8.826311111450195, "y": 10.612848281860352, "z": -8.794543266296387};
      const r2Pt = { "x": -1.5316486358642578, "y": -17.590429306030273, "z": -3.607043504714966}
      let cpPt = JSON.parse(JSON.stringify(pPt));
      let mvPt = .005;
      let cnt = 0;


      self.Robot1(r1Pt.x, r1Pt.y, r1Pt.z);
      self.Robot2(r2Pt.x, r2Pt.y, r2Pt.z);
      self.Line3(r2Pt.x, r2Pt.y, r2Pt.z, r1Pt.x, r1Pt.y, r1Pt.z);

      function doStuff () {
        cnt = cnt + 1;
        mvPt = mvPt * -1;
        cpPt.x = cpPt.x + mvPt;
        cpPt.y = cpPt.y + mvPt;
        if(cnt === 20)
        {
          if(mvPt === .5 || mvPt === -.5)
          {mvPt = mvPt / 100}
          else{ mvPt = mvPt * 10}
          cnt = 0;
        }
        self.Line1(r1Pt.x, r1Pt.y, r1Pt.z, cpPt.x, cpPt.y, cpPt.z);
        self.Line2(r2Pt.x, r2Pt.y, r2Pt.z, cpPt.x, cpPt.y, cpPt.z);
        self.Prism(cpPt.x, cpPt.y, cpPt.z);
      }

      setInterval(  () => {doStuff();},  250)
    }

    fakePoints(){
      let i = 0;
      let MAX_POINTS = 5000;
      while (i < MAX_POINTS) {
        var x = 200*Math.random();
        var y = 200*Math.random();
        var z = 200*Math.random();
        APLpts.push({PointNumber: i, position: new THREE.Vector3(x,y,z)});
        i++;
      }
      self.PointsAdd(APLpts);
      document.addEventListener( 'mousemove', self.onDocumentMouseMove, false );
    }

    onDocumentMouseMove( event ) {
      event.preventDefault();
      let pointerVector = new THREE.Vector3()
      let pointerDir = new THREE.Vector3()
      let rect = self.viewer.impl.canvas.getBoundingClientRect();

      self.mouse.x = ((event.clientX - rect.left) / rect.width) * 2 - 1;
      self.mouse.y = -((event.clientY - rect.top) / rect.height) * 2 + 1;

      if (self.camera.isPerspective) {
        pointerVector.set(self.mouse.x, self.mouse.y, 0.5);
        pointerVector.unproject(self.camera);
        self.raycaster.set(self.camera.position,pointerVector.sub( self.camera.position).normalize());
      } else {
        pointerVector.set(self.mouse.x, self.mouse.y, -1);
        pointerVector.unproject(self.camera);
        pointerDir.set(0, 0, -1);
        self.raycaster.set(pointerVector,pointerDir.transformDirection(self.camera.matrixWorld));
      }
      //let objs = window.b3l_pts.map(a => a.threeobj);
      //let intersections = self.raycaster.intersectObjects(objs, true);
      let intersections = self.raycaster.intersectObjects(window.b3l_pcloud, true);
            if ( intersections.length > 0 ) {
                // if ( self.intersected != intersections[ 0 ].object ) {
          //if ( self.intersected ) self.intersected.material.color.setHex( pointColor );
          self.highlightObj.position.copy( intersections[0].point);
          self.highlightObj.visible = true;
          self.viewer.impl.invalidate(false, false, true);
                    // self.intersected = intersections[ 0 ].object;
                    // self.intersected.material.color.setHex( intersectColor );
                // }
            }
            else if ( self.highlightObj.visible ) {//self.intersected ) {
        // self.intersected.material.color.setHex( pointColor );
        self.highlightObj.visible = false;
        self.viewer.impl.invalidate(false, false, true);
                // self.intersected = null;
            }
    }

    hideObjectsBlockingObject(obj){
      let camera = self.viewer.navigation.getCamera();
      // let position = this.getPosition(this._viewer.canvas);
      // let pt = this.getScreenPt(event, position);
      // Rebase both X and Y in -1 to 1 (relative to the screen center and size)
      // let x = (pt.x / position.w) * 2 - 1;
      // let y = -(pt.y / position.h) * 2 + 1;
      // let pointerVector = new THREE.Vector3();
      // let pointerDir = new THREE.Vector3();
      // let origin, vector;
      // if (camera.isPerspective) {
      //     pointerVector.set(x, y, 0.5);
      //     pointerVector.unproject(camera);
      let pointerVector = obj.position.clone();
      let origin = camera.position;
      let vector = pointerVector.sub(camera.position).normalize();
      // } else {
      //     pointerVector.set(x, y, -1);
      //     pointerVector.unproject(camera);
      //     pointerDir.set(0, 0, -1);
      //     origin = pointerVector;
      //     vector = pointerDir.transformDirection(camera.matrixWorld);
      // }
      let intersects = self.getIntersections(origin, vector, self.distanceBetween(origin,obj.position));
      if(intersects.length > 0){
        let dbIds = intersects.map(a => a.dbId);
        self.viewer.hide(dbIds);
      }
    }

    hideObjectsAlongVector(pt1, pt2){
      let vector = new THREE.Vector3();
      vector.subVectors( pt2, pt1).normalize();
      let intersects = self.getIntersections(pt1, vector, self.distanceBetween(pt1,pt2));
      if(intersects.length > 0){
        let dbIds = intersects.map(a => a.dbId);
        self.viewer.hide(dbIds);
      }
    }

    getIntersections(origin, vector, far) {
      let intersects = [];
      let raycaster = new THREE.Ray(origin, vector, 0, far);
      // Fire the ray
      this.viewer.impl.rayIntersect(raycaster, true, false, false, intersects);
      return intersects;
    }

    distanceBetween( v1, v2 ) //when Three updated use distanceTo instead
    {
        let dx = v1.x - v2.x;
        let dy = v1.y - v2.y;
        let dz = v1.z - v2.z;
        return Math.sqrt( dx * dx + dy * dy + dz * dz );
    }

    CreatePointFromItmProp(itm){
      return new Promise(function(resolve, reject) {
      if(itm.properties['PointNumber'] !== '' || itm.properties['POINTNUMBER'] !== '' ){
        //lAPLpts.push(itm['dbId']);   //error, not defined
        let pt = {};
        pt.dbid = itm['dbId'];
        try{pt.PointNumber = itm.properties.find((props)=>{return props.attributeName === 'PointNumber';}).displayValue;} catch(e){pt.PointNumber = ''};
        if(pt.PointNumber === ''){
          try{pt.PointNumber = itm.properties.find((props)=>{return props.attributeName === 'POINTNUMBER';}).displayValue;} catch(e){pt.PointNumber = ''};
        }
        if(pt.PointNumber === ''){
          try{pt.PointNumber = itm.properties.find((props)=>{return props.attributeName === 'LcOaPropOverrideCat:PointNumber';}).displayValue;} catch(e){pt.PointNumber = ''};
        }
        if(pt.PointNumber === ''){
          try{pt.PointNumber = itm.properties.find((props)=>{return props.displayName === 'PointNumber';}).displayValue;} catch(e){pt.PointNumber = ''};
        }
        if(pt.PointNumber === ''){
          try{pt.PointNumber = itm.properties.find((props)=>{return props.displayName === 'POINTNUMBER';}).displayValue;} catch(e){pt.PointNumber = ''};
        }

        if(pt.PointNumber !== ''){
          try{pt.PointDescription = itm.properties.find((props)=>{return props.attributeName === 'PointDescription';}).displayValue;} catch(e){pt.PointDescription = ''};
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.attributeName === 'DESCRIPTION';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.attributeName === 'DESC';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.attributeName === 'LcOaPropOverrideCat:DESC';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.displayName === 'PointDescription';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.displayName === 'Description';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.displayName === 'DESCRIPTION';}).displayValue;} catch(e){pt.PointDescription = ''};
          }
          if(pt.PointDescription === ''){
            try{pt.PointDescription = itm.properties.find((props)=>{return props.displayName === 'DESC';}).displayValue;} catch(e){pt.PointDescription = ''};
          }

          try{pt.PointRole = itm.properties.find((props)=>{return props.attributeName === 'PointRole';}).displayValue;} catch(e){pt.PointRole = ''};
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.attributeName === 'POINTROLE';}).displayValue;} catch(e){pt.PointRole = ''};
          }
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.attributeName === 'Role';}).displayValue;} catch(e){pt.PointRole = ''};
          }
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.attributeName === 'LcOaPropOverrideCat:Role';}).displayValue;} catch(e){pt.PointRole = ''};
          }
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.displayName === 'PointRole';}).displayValue;} catch(e){pt.PointRole = ''};
          }
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.displayName === 'POINTROLE';}).displayValue;} catch(e){pt.PointRole = ''};
          }
          if(pt.PointRole === ''){
            try{pt.PointRole = itm.properties.find((props)=>{return props.displayName === 'Role';}).displayValue;} catch(e){pt.PointRole = ''};
          }

          try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'PointType';}).displayValue;} catch(e){pt.PointType = ''};
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'POINTTYPE';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.displayName === 'PointType';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.displayName === 'POINTTYPE';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'LcOaPropOverrideCat:PointType';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'ALL_MODEL_TYPE_NAME';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'LcRevitData_Element:LcRevitPropertyElementType';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.displayName === 'SYSTEM';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'Type';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.displayName === 'Type';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.attributeName === 'Type Name';}).displayValue;} catch(e){pt.PointType = ''};
          }
          if(pt.PointType === ''){
            try{pt.PointType = itm.properties.find((props)=>{return props.displayName === 'Type Name';}).displayValue;} catch(e){pt.PointType = ''};
          }
          self.lAPLpts.push(pt);
        }
      }
      resolve();
      });
    }

    findLocalPoints(){
      self.getAllLeafComponents(self.viewer, function (allDbIds) {
      //self.getAllTreeComponents(self.viewer, function (allDbIds) {
        var usedDbIds = [];
        self.viewer.model.getBulkProperties(allDbIds, fullPropFilter, (result) => {
          if (result.length > 0) {
            result.forEach((itm) => {
              usedDbIds.push(itm.dbId);
              self.CreatePointFromItmProp(itm);
            });
          }
          //Now get a list of leaf dbids that did not have the above properties found and check their parent instead
          var chldDbIds = allDbIds.filter( ( itm ) => !usedDbIds.includes( itm ) );
          //get a list of parent dbids, only add if not already in the list
          self.viewer.model.getBulkProperties(chldDbIds, ['parent[dbid]'], (result) => {
            var parDbIds = [];
            if (result.length > 0) {
              result.forEach((itm) => {
                var tmpId;
                try{tmpId = itm.properties.find((props)=>{return props.attributeName === 'parent[dbid]';}).displayValue;} catch(e){};
                if(tmpId !== ''){
                  if(parDbIds.indexOf(tmpId) === -1){parDbIds.push(tmpId)};
                }
              });
            }
            //Now check for the properties on the list of parents
            usedDbIds = [];
            if(parDbIds.length > 0){
              self.viewer.model.getBulkProperties(parDbIds, fullPropFilter, (result) => {
                if (result.length > 0) {
                  result.forEach((itm) => {
                    usedDbIds.push(itm.dbId);
                    self.CreatePointFromItmProp(itm);
                  });
                }
                //need to check one more level up
                //Now get a list of leaf dbids that did not have the above properties found and check their parent instead
                var chldDbIds = parDbIds.filter( ( itm ) => !usedDbIds.includes( itm ) );
                //get a list of parent dbids, only add if not already in the list
                self.viewer.model.getBulkProperties(chldDbIds, ['parent[dbid]'], (result) => {
                  var par2DbIds = [];
                  if (result.length > 0) {
                    result.forEach((itm) => {
                      var tmpId;
                      try{tmpId = itm.properties.find((props)=>{return props.attributeName === 'parent[dbid]';}).displayValue;} catch(e){};
                      if(tmpId !== ''){
                        if(par2DbIds.indexOf(tmpId) === -1){par2DbIds.push(tmpId)};
                      }
                    });
                  }
                  //Now check for the properties on the list of parents
                  usedDbIds = [];
                  if(par2DbIds.length > 0){
                    self.viewer.model.getBulkProperties(par2DbIds, fullPropFilter, (result) => {
                      if (result.length > 0) {
                        result.forEach((itm) => {
                          usedDbIds.push(itm.dbId);
                          self.CreatePointFromItmProp(itm);
                        });
                      }
                      self.setCoordsToPoints();
                    })
                  }else{
                    self.setCoordsToPoints();
                  }
                })
              })
            }else{
              self.setCoordsToPoints();
            }
          })
        })
      })
    }

    setVisibilityOn(itm) {
      return Promise.resolve(self.viewer.impl.visibilityManager.setNodeOff(itm.dbId, false));
    }
    setVisibilityOff(itm) {
      return Promise.resolve(self.viewer.impl.visibilityManager.setNodeOff(itm.dbId, true));
    }
    
    modifyLocalPoints(modFunction, finalFunction, propFilter) {
      self.getAllLeafComponents(self.viewer, function (allDbIds) {
        const it = self.viewer.model.getInstanceTree();
        function checkProperties ( dbIds, callback) {
          if( !dbIds || dbIds.length === 0){
              callback();
          }else{
            var usedDbIds = [];
            var fn = function actionItm(itm){
              return new Promise(function(resolve, reject) {
                usedDbIds.push(itm.dbId);
                modFunction(itm).then(()=>{resolve()});
              })
            }
            Promise.resolve(self.viewer.model.getBulkProperties(dbIds, propFilter, (results) => {
              var actions = results.map(fn);
              Promise.all(actions).then(()=>{
                var leftDbIds = dbIds.filter( ( dbId ) => !usedDbIds.includes( dbId ) );
                if(leftDbIds.length > 0)
                  checkProperties(getParIds(leftDbIds), callback);
                else
                  callback();
              });
            }))
          }
         }

        function getParIds(dbIds) {
          //get a list of dbids that did not have the properties found and find their parent to check instead
          let parDbIds = [];
          dbIds.forEach((dbId) => {
            let tmpId = it.getNodeParentId(dbId);
            if(tmpId !== '' && tmpId !== 0){
              if(parDbIds.indexOf(tmpId) === -1){parDbIds.push(tmpId)};
            }
          })
          return parDbIds;
        }

        checkProperties(allDbIds, ()=> {
          if(finalFunction)
            finalFunction();
        });
      })
    }

    hideLocalPoints(){
      self.modifyLocalPoints(self.setVisibilityOff, undefined, partPropFilter);
    };

    showLocalPoints() {
      self.modifyLocalPoints(self.setVisibilityOn, undefined, partPropFilter);
    };
    
    setCoordsToPoints(){
      self.lAPLpts.forEach((ptObj)=>{
        //var xyz = {x:0,y:0,z:0};
        var point = new THREE.Vector3 (0,0,0);
        var fragId = self.getFragIdFromDbId(self.viewer, ptObj.dbid);
        if(fragId && fragId !== 0){
          var mesh = self.viewer.impl.getRenderProxy(self.viewer.model, fragId);
          var matrixWorld = mesh.matrixWorld;
          point = new THREE.Vector3 (matrixWorld.elements[12],matrixWorld.elements[13],matrixWorld.elements[14])
        }else{
          var bBox = self.getOriginalWorldBoundingBoxFromDbId(self.viewer, ptObj.dbid);
          if (bBox)
          {
            point = new THREE.Vector3 ((bBox.max.x + bBox.min.x)/2,(bBox.max.y + bBox.min.y)/2,(bBox.max.z + bBox.min.z)/2)
          }
        }
        ptObj.position = point;
      })
      console.log("INFO: LOCALPOINTS: FindLocalPoints 4 : Count: " + self.lAPLpts.length);
      self.LocalPtsFound = 2
    }

    resetLocalPoints(){
      self.LocalPtsFound = 0
    }

    returnLocalPoints(){
      if(self.LocalPtsFound === 2){
        console.log("INFO: LOCALPOINTS 2: findLocalPoints completed, returning found points");
        return self.lAPLpts;
      }else if (self.LocalPtsFound === 1){
        console.log("INFO: LOCALPOINTS 1: findLocalPoints still processing");
        return '1';
      }else if (self.LocalPtsFound === 0){
        console.log("INFO: LOCALPOINTS 0: findLocalPoints triggered");
        //self.findLocalPoints();
        self.modifyLocalPoints(self.CreatePointFromItmProp, self.setCoordsToPoints, fullPropFilter);
        self.LocalPtsFound = 1;
        return '1';
      }
    }

    markLocalPoints(){
      if (self.APLpts.length < 1){
        this.findLocalPoints();
        }        
      if (self.APLpts.length > 0){
        self.PointsAdd(self.APLpts);
        document.addEventListener( 'mousemove', self.onDocumentMouseMove, false );
      }
    }

    getAllTreeComponents(viewer, callback, base) {
       var dbIds = Object.keys(viewer.model.getData().instanceTree.nodeAccess.dbIdToIndex);
       callback(dbIds);
    }
    getAllLeafComponents(viewer, callback, base) {
      var cbCount = 0; // count pending callbacks
      var components = []; // store the results
      var tree; // the instance tree
      function getLeafComponentsRec(parent) {
          cbCount++;
          if (tree.getChildCount(parent) != 0) {
              tree.enumNodeChildren(parent, function (children) {
                  getLeafComponentsRec(children);
              }, false);
          } else {
              components.push(parent);
          }
          if (--cbCount == 0) callback(components);;
      }
      viewer.getObjectTree(function (objectTree) {
          tree = objectTree;
          base = base || tree.getRootId();
          var allLeafComponents = getLeafComponentsRec(base);
      });
    }

    getOriginalWorldBoundingBoxFromDbId(viewer, dbid)
    {
      //First check if its a leaf and if not, then find all leafs it contains
      var dbIds = [];
      self.getAllLeafComponents(viewer, function (allDbIds) {dbIds = allDbIds;}, dbid);
      //create a list of all fragids for the dbid or dbid's
      var fragIds = []; // store the results
      dbIds.forEach((dbi)=>{
        var frags = this.getLeafFragIds(viewer,dbi);
        if (frags.length > 0 && !(frags.length === 1 && frags[0] === 0)) {
          fragIds.push(frags)
        }
      })
      //get a bounding box for all the related fragids
      if(fragIds.length > 0){
        var bBox = self.getOriginalWorldBoundingBox(fragIds,viewer.model.getFragmentList());
        return bBox;
      }else{
        return null;
      }
    }

    getLeafFragIds( viewer, leafId ) {
      const instanceTree = viewer.model.getData().instanceTree;
      const fragIds = [];
      instanceTree.enumNodeFragments( leafId, function( fragId ) {
        fragIds.push( fragId );
      });
      return fragIds;
    }

    getFragIdFromDbId(viewer, dbid) {
      var returnValue;
      var it = viewer.model.getData().instanceTree;
      it.enumNodeFragments(dbid, function (fragId) {
        returnValue = fragId;
      }, false);
      return returnValue;
    }

    // Returns bounding box as loaded in the file
    // (no explosion nor transformation)
    getOriginalWorldBoundingBox(fragIds, fragList) {

      var fragBoundingBox = new THREE.Box3();
      var nodeBoundingBox = new THREE.Box3();

      var fragmentBoxes = fragList.boxes;

      fragIds.forEach(function(fragId) {

          var boffset = fragId * 6;

          fragBoundingBox.min.x = fragmentBoxes[boffset];
          fragBoundingBox.min.y = fragmentBoxes[boffset+1];
          fragBoundingBox.min.z = fragmentBoxes[boffset+2];
          fragBoundingBox.max.x = fragmentBoxes[boffset+3];
          fragBoundingBox.max.y = fragmentBoxes[boffset+4];
          fragBoundingBox.max.z = fragmentBoxes[boffset+5];

          nodeBoundingBox.union(fragBoundingBox);
      });
      return nodeBoundingBox;
    }

    LoadObjectJson(name){
      // instantiate a loader
      var loader = new THREE.ObjectLoader();
      var geometryPromise = new Promise(function(resolve, reject) {
        // load a resource
        loader.load('/extensions/Viewing.Extension.B3L/Objects/'+name+'.json', function ( obj ) {
          resolve(obj);
        });
      });
      return geometryPromise;
    }

    GetSphere(color, x, y, z)
    {
      var sphere = new THREE.Mesh(new THREE.SphereGeometry(.1), new THREE.MeshBasicMaterial({
        color: color
      }));
      sphere.position.set(x, y, z);
      return sphere;
    }

    GetObject(name,color,x,y,z)
    {
      let p1 = new Promise((resolve) =>{
        if(UseObjects){
          self.LoadObjectJson(name)
          .then((object)=>{
                //Set scale, expect our objects to be in meters
                var scale = 1.0 / self.viewer.model.getUnitScale();
                object.scale.set(scale,scale,scale)
                object.position.set(x, y, z);
                resolve(object);
              }
            );
        }else{
          let sphere = self.GetSphere(color, x, y, z);
          resolve(sphere);
        }
      });
      return p1;
    }

    PrismVis(show = false) {
      if (window.b3l_prism) {
        window.b3l_prism.visible = show;
      }
    }
    Prism2Vis(show = false) {
      if (window.b3l_prism2) {
        window.b3l_prism2.visible = show;
      }
    }

    Prism(x,y,z,model = 'Prism2') {
      if (!window.b3l_prism) {
        self.GetObject(model,'red',x,y,z).then(function(object){
          window.b3l_prism = object;
          self.viewer.impl.createOverlayScene('B3LPRISM', null, null);
          self.viewer.impl.addOverlay('B3LPRISM', window.b3l_prism);
        });
      } else {
        window.b3l_prism.position.set(x, y, z);
        self.viewer.impl.invalidate(false, false, true);
      }
    }

    Prism2(x,y,z, model = 'GPS1') {
      if (!window.b3l_prism2) {
        self.GetObject(model,'red',x,y,z).then(function(object){
          window.b3l_prism2 = object;
          self.viewer.impl.createOverlayScene('B3LPRISM2', null, null);
          self.viewer.impl.addOverlay('B3LPRISM2', window.b3l_prism2);
        });
      } else {
        window.b3l_prism2.position.set(x, y, z);
        self.viewer.impl.invalidate(false, false, true);
      }
    }

    Robot1Vis(show = false) {
      if (window.b3l_robot1) {
        window.b3l_robot1.visible = show;
      }
    }
    Robot2Vis(show = false) {
      if (window.b3l_robot2) {
        window.b3l_robot2.visible = show;
      }
    }
    Robot1(x,y,z) {
      if (!window.b3l_robot1) {
        self.GetObject('Robot1','green',x,y,z).then(function(object){
        window.b3l_robot1 = object;
        self.viewer.impl.createOverlayScene('B3LROBOT1', null, null);
        self.viewer.impl.addOverlay('B3LROBOT1', window.b3l_robot1);
        });
      } else {
        window.b3l_robot1.position.set(x, y, z);
        self.viewer.impl.invalidate(false, false, true);
      }
    }

    Robot2(x,y,z) {
      if (!window.b3l_robot2) {
        self.GetObject('Robot2','blue',x,y,z).then(function(object){
          window.b3l_robot2 = object;
          self.viewer.impl.createOverlayScene('B3LROBOT2', null, null);
          self.viewer.impl.addOverlay('B3LROBOT2', window.b3l_robot2);
        });
      } else {
        window.b3l_robot2.position.set(x, y, z);
        self.viewer.impl.invalidate(false, false, true);
      }
    }

    Line1(x1,y1,z1,x2,y2,z2,color = 0xff5900){
      self.Line(x1,y1,z1,x2,y2,z2,color,window.b3l_line1,'b3l_line1',true);
    }

    Line2(x1,y1,z1,x2,y2,z2,color = 0x0000FF){
      self.Line(x1,y1,z1,x2,y2,z2,color,window.b3l_line2,'b3l_line2',true);
    }

    Line3(x1,y1,z1,x2,y2,z2,color = 0x4B9609){
      self.Line(x1,y1,z1,x2,y2,z2,color,window.b3l_line3,'b3l_line3',false);
    }

    Line4(x1,y1,z1,x2,y2,z2,color = 0x4B9609){
      self.Line(x1,y1,z1,x2,y2,z2,color,window.b3l_line4,'b3l_line4',true);
    }

    Line(x1,y1,z1,x2,y2,z2,color,windowParam,sceneName,dashed){
      if (!windowParam) {
        var geometry = new THREE.Geometry();
        geometry.vertices.push(new THREE.Vector3(x1,y1,z1));
        geometry.vertices.push(new THREE.Vector3(x2,y2,z2));
        geometry.computeLineDistances();
        if (dashed)
            var material = new THREE.LineDashedMaterial({ color: color, dashSize: 1, gapSize: 1, linewidth: 2 });
        else
            var material = new THREE.LineBasicMaterial({ color: color, linewidth: 2 });
        material.depthTest = false;
        material.depthWrite = false;
    
        var object = new THREE.Line(geometry, material);
        object.geometry.dynamic = true;
        windowParam = object;
        self.viewer.impl.createOverlayScene(sceneName, null, null);
        self.viewer.impl.addOverlay(sceneName, windowParam);
      } else {
        windowParam.geometry.vertices[0].position = new THREE.Vector3(x1,y1,z1);
        windowParam.geometry.vertices[1].position = new THREE.Vector3(x2,y2,z2);
        windowParam.geometry.verticesNeedUpdate = true;
        self.viewer.impl.invalidate(false, false, true);
      }
    }

    PointsAdd(ptArray){
      for ( var i = 0; i < ptArray.length; i ++ ) {
        self.PointAdd(ptArray[i].PointNumber,ptArray[i].position.x,ptArray[i].position.y,ptArray[i].position.z, false);
      }
      self.viewer.impl.invalidate(false, false, true);      
    }
    
    PointAdd(id,x,y,z,redraw = true){
      if (!window.b3l_pcloud) {
        self.CreatePcloud();
      }
      if(!window.b3l_pts[id]){
        window.b3l_pts[id] = {};
        window.b3l_pts[id].position = [x,y,z];
        window.b3l_pts[id].threeobj = self.GetSphere(pointColor, x, y, z);
        self.viewer.impl.addOverlay('B3LPCLOUD', window.b3l_pts[id].threeobj);
        // let length = Object.keys(window.b3l_pts).length;
        // let positions = window.b3l_pcloud.geometry.attributes.position.array;
        // let index = length - 1 * 3;
        // positions[ index ++ ] = x;
        // positions[ index ++ ] = y;
        // positions[ index ++ ] = z;
        // window.b3l_pcloud.geometry.setDrawRange( 0, length);
        // window.b3l_pcloud.geometry.attributes.position.needsUpdate = true;
        ////window.b3l_pcloud.geometry.vertices.push(new THREE.Vector3(x,y,z));
        ////window.b3l_pcloud.geometry.vertices.needsUpdate = true;
        window.b3l_pcloud.push(window.b3l_pts[id].threeobj);
        if(redraw){
          self.viewer.impl.invalidate(false, false, true);          
        }
      }
    }
    
    PointRemove(id,redraw = true){
      if (!window.b3l_pcloud) {
        CreatePcloud();
      }
      if(window.b3l_pts[id]){

      }
      window.b3l_pcloud.geometry.setDrawRange( 0, Object.keys(window.b3l_pts).length);
      window.b3l_pcloud.geometry.attributes.position.needsUpdate = true;
      if(redraw){
        self.viewer.impl.invalidate(false, false, true);          
      }
  }
    
    PointsShow(){
      if (!window.b3l_pcloud) {
        self.CreatePcloud();
      }
      window.b3l_pcloud.visible = true;
    }
    
    PointsHide(){
      if (!window.b3l_pcloud) {
        self.CreatePcloud();
      }
      window.b3l_pcloud.visible = false;
    }

    CreatePcloud()
    {
      if (!window.b3l_pcloud) {
        //Cannot use Points object or even setDrawRange as they do not exist in the R71 version of threeJS used by LMV :(
        // let MAX_POINTS = 100000;
        // let geometry = new THREE.BufferGeometry();
        // let positions = new Float32Array( MAX_POINTS * 3 ); 
        // geometry.addAttribute( 'position', new THREE.BufferAttribute( positions, 3 ) );
        //geometry.addAttribute( 'color', new THREE.BufferAttribute( colors, 3 ).setDynamic( true ) );
        //let material = new THREE.PointsMaterial({color: 0xff0000,size: 2,opacity: 1});
        let geometry = new THREE.Geometry();
        let material = new THREE.PointCloudMaterial({color: pointColor,size: 5,sizeAttenuation: false ,opacity: 1});
        window.b3l_pts = {};
        //window.b3l_pcloud = new THREE.Points(geometry, material);
        ////window.b3l_pcloud = new THREE.PointCloud(geometry, material);
        //window.b3l_pcloud.geometry.setDrawRange( 0, 0);
        window.b3l_pcloud = [];
        self.viewer.impl.createOverlayScene('B3LPCLOUD', null, null);
        ////self.viewer.impl.addOverlay('B3LPCLOUD', window.b3l_pcloud);
        self.viewer.impl.addOverlay('B3LPCLOUD', self.highlightObj);
        self.viewer.impl.addOverlay('B3LPCLOUD', self.selectObj);
      }
    }

    toggleSection()
    {
        if(!sectionActive && sectionPlanes.length > 0)
        {
          self.activateSection();
        }
        else
        {
          self.deactivateSection();
        }
    }

    deactivateSection()
    {
      if(sectionActive)
      {
        self.viewer.removeEventListener(Autodesk.Viewing.CUTPLANES_CHANGE_EVENT, this.checkSection);
        self.viewer.impl.setCutPlaneSet("B3Layout")
        sectionActive = false;
      }
    }

    activateSection()
    {
      if(sectionPlanes.length > 0)
      {
        self.viewer.impl.setCutPlaneSet("B3Layout", sectionPlanes)
        self.viewer.addEventListener(Autodesk.Viewing.CUTPLANES_CHANGE_EVENT, this.checkSection);
        sectionActive = true;
      }
    }

    checkSection()
    {
      if(sectionActive === true && sectionPlanes.length > 0)
      {
        var cpSets = self.viewer.impl.getCutPlaneSets();
        for (var i=0; i<cpSets.length; i++) {
          var cpName = cpSets[i];
          if (cpName === "B3Layout") {
            if(self.viewer.impl.getCutPlanes().length < sectionPlanes.length)
              self.viewer.impl.setCutPlaneSet("B3Layout", sectionPlanes)
          }else{
            continue;
          }
        }
      }
    }

    setSectionOnSelection(distance, height, depth)
    {
      let box = new THREE.Box3();
      let sels = self.viewer.getSelection();
      let center = undefined;
      if(sels.length > 0){
        let fragids = [];
        self.viewer.model.getData().instanceTree.enumNodeFragments(self.viewer.getSelection()[0], function(fragId){fragids.push(fragId)});
        let fraglst = self.viewer.model.getFragmentList();
        if(fragids.length > 0){
          fraglst.getWorldBounds(fragids[0], box);
          if(box)
          center = box.center();
        }
      }
      self.setSection(center,distance, height, depth);
    }

    setSection(origin, distance, height, depth)
    {
      let normals = [
          new THREE.Vector3(1, 0, 0),
          new THREE.Vector3(0, 1, 0),
          new THREE.Vector3(0, 0, 1),
          new THREE.Vector3(-1, 0, 0),
          new THREE.Vector3(0, -1, 0),
          new THREE.Vector3(0, 0, -1)
      ];

      let obbox = self.viewer.impl.getVisibleBounds();
      let center = (origin !== undefined) ? origin : obbox.center();
      height = (height !== undefined && height > 0) ? height : 9;
      depth = (depth !== undefined && depth > 0) ? depth : 2;
      distance = (distance !== undefined) ? distance : 100;
      let distances = [
        distance + center.x,
        distance + center.y,
        height+ center.z,
        distance - center.x,
        distance - center.y,
        depth-center.z
      ];
      sectionPlanes = [];
      for (var i = 0; i < normals.length; i++) {
        sectionPlanes.push(new THREE.Vector4(normals[i].x, normals[i].y, normals[i].z, -1 * distances[i]));
      }
      self.activateSection();
    }

    logProperties(parent, parent2)
    {
        if(self.viewer.getSelection()[0])
        {
            let dbid = self.viewer.getSelection()[0];
            if (parent === true)
                dbid = self.viewer.model.getData().instanceTree.getNodeParentId(dbid);
            if (parent2 === true)
                dbid = self.viewer.model.getData().instanceTree.getNodeParentId(dbid);
            self.viewer.getProperties(dbid, function(e) {
                for (var t = (e.name,
                e.properties), n = {}, r = 0, i = 0; i < t.length; i += 1) {
                    var o = t[i];
                    if (!o.hidden) {
                        var a = o.displayCategory
                          , s = o.displayName
                          , r = o.attributeName
                          , u = Autodesk.Viewing.Private.formatValueWithUnits(o.displayValue, o.units, o.type);
                        n[a] || (n[a] = {},
                        n[a].children = [],
                        n[a].order = r,
                        r += 1);
                        var l = {
                            name: s,
                            attName: r,
                            value: u
                        };
                        n[a].children.push(l)
                    }
                }
                console.log(JSON.stringify(n))
            })
          }
    }
  }
  Autodesk.Viewing.theExtensionManager.registerExtension(ExtensionId, B3Layout)
}